﻿using System;
using System.Linq;
using System.Security.Cryptography;

namespace HeuristicLab.Algorithms.DataAnalysis.SymRegGrammarEnumeration {
  public static class HashExtensions {
    public sealed class Node : IComparable<Node>, IEquatable<Node> {
      public Symbol Symbol;
      public int Arity;
      public bool Enabled;
      public int Hash;
      public int Size;

      public bool IsChild => Arity == 0;

      private Node(Symbol s, int n, bool e) {
        Symbol = s;
        Arity = Size = n;
        Enabled = e;
        Hash = Symbol.GetHashCode();
      }

      public static Node Make(Symbol s, int n, bool e = true) { return new Node(s, n, e); }

      public int CompareTo(Node other) {
        var res = Symbol.CompareTo(other.Symbol);
        return res == 0 ? Hash.CompareTo(other.Hash) : res;
      }

      public override string ToString() {
        return $"{Symbol} {Arity} {Size} {Hash} {Enabled}";
      }

      public bool Equals(Node other) {
        return Symbol.Equals(other.Symbol) && Hash.Equals(other.Hash);
      }

      public override bool Equals(object obj) {
        var other = obj as Node;
        if (other != null)
          return Equals(other);
        return base.Equals(obj);
      }

      public override int GetHashCode() {
        return Hash;
      }

      public static bool operator ==(Node a, Node b) {
        return a.Equals(b);
      }

      public static bool operator !=(Node a, Node b) {
        return !(a == b);
      }
    }

    private static int Arity(this Grammar g, Symbol s) {
      if (s == g.Addition || s == g.Multiplication)
        return 2;

      if (s == g.Log || s == g.Exp || s == g.Sin || s == g.Inv)
        return 1;

      return 0;
    }

    public static int ComputeHash(this Grammar grammar, HashAlgorithm ha, SymbolList sentence) {
      var nodes = sentence.Select(x => Node.Make(x, grammar.Arity(x))).ToArray();
      var simplified = grammar.Simplify(ha, nodes);

      int hash = 1315423911;
      foreach (var node in simplified)
        hash ^= (hash << 5) + node.Hash + (hash >> 2);
      return hash;

      //var hashes = simplified.Select(x => x.Hash).ToArray();
      //return BitConverter.ToInt32(ComputeHash(ha, hashes), 0);
    }

    public static SymbolList Simplify(this Grammar grammar, HashAlgorithm ha, SymbolList sentence) {
      var nodes = sentence.Select(x => Node.Make(x, grammar.Arity(x))).ToArray();
      return new SymbolList(grammar.Simplify(ha, nodes).Select(x => x.Symbol));
    }

    private static Node[] Simplify(this Grammar grammar, HashAlgorithm ha, Node[] nodes) {
      var reduced = grammar.Reduce(nodes);
      SortChildren(ha, reduced);

      for (int i = 0; i < reduced.Length; ++i) {
        var node = reduced[i];
        if (node.IsChild) {
          continue;
        }

        if (node.Symbol == grammar.Addition) {
          grammar.SimplifyAddition(reduced, i);
        } else if (node.Symbol == grammar.Multiplication) {
          grammar.SimplifyMultiplication(reduced, i);
        }
      }
      // detect if anything was simplified
      var count = 0;
      foreach (var node in reduced) {
        if (!node.Enabled) { ++count; }
      }
      if (count == 0) {
        return reduced;
      }

      var simplified = new Node[reduced.Length - count];
      int j = 0;
      foreach (var node in reduced) {
        if (node.Enabled) {
          simplified[j++] = node;
        }
      }
      UpdateNodeSizes(simplified);
      SortChildren(ha, simplified);
      return simplified;
    }

    /// <summary>
    /// This method will set @Node @Enabled flags to false for nodes that should be simplified
    /// </summary>
    /// <param name="nodes">The array of nodes representing the expression</param>
    /// <param name="i">Index of the current node</param>
    private static void SimplifyAddition(this Grammar grammar, Node[] nodes, int i) {
      // simplify additions of terms by eliminating terms with the same symbol and hash
      var children = IterateChildren(nodes, i);

      var curr = children[0];
      var node = nodes[i];

      for (int j = 1; j < children.Length; ++j) {
        var idx = children[j];
        if (nodes[idx] == nodes[curr]) {
          for (int k = idx - nodes[idx].Size; k <= idx; ++k) {
            nodes[k].Enabled = false;
          }
          node.Arity--;
        } else {
          curr = idx;
        }
      }
      if (node.Arity == 1) { // if the arity is 1 we don't need the addition node at all
        node.Enabled = false;
      }
    }

    private static void SimplifyMultiplication(this Grammar grammar, Node[] nodes, int i) {
      // simplify multiplifactions by reducing inv terms for example a * 1/a -> c
      var node = nodes[i];
      var children = IterateChildren(nodes, i);
      foreach (var c in children) {
        if (nodes[c].Symbol == grammar.Inv) {
          var denominator = nodes[c - 1];
          foreach (var d in children) {
            if (nodes[d].Enabled && nodes[d] == denominator) {
              // the inv node can be simplified because the denominator matches another child
              // disable the inv node, its child, and the matching child
              nodes[d].Enabled = false;
              nodes[c].Enabled = false;
              nodes[c - 1].Enabled = false;
              node.Arity -= 2;
              break;
            }
          }

          if (node.Arity == 0) { // if everything is simplified this node becomes constant
            nodes[i] = Node.Make(grammar.Const, 0);
          } else if (node.Arity == 1) { // when i have only 1 arg left i can skip this node
            node.Enabled = false;
          }
        }
      }
    }

    private static int ComputeHash(HashAlgorithm ha, Node[] nodes, int i) {
      var node = nodes[i];
      var hashes = new int[node.Arity + 1];
      int idx = 0;
      foreach (int c in IterateChildren(nodes, i)) {
        hashes[idx++] = nodes[c].Hash;
      }
      hashes[idx] = node.Symbol.GetHashCode();
      //return BitConverter.ToInt32(ComputeHash(ha, hashes), 0);
      return HashUtil.JSHash(hashes);
    }

    private static void SortChildren(HashAlgorithm ha, Node[] nodes) {
      for (int i = 0; i < nodes.Length; ++i) {
        var node = nodes[i];

        if (node.IsChild) {
          continue;
        }

        var arity = node.Arity;
        var size = node.Size;

        if (arity == size) { // all child nodes are terminals
          Array.Sort(nodes, i - size, size);
        } else { // i have some non-terminal children
          var sorted = new Node[size];
          var indices = IterateChildren(nodes, i);
          Array.Sort(indices, (a, b) => nodes[a].CompareTo(nodes[b]));

          int idx = 0;
          foreach (var j in indices) {
            var child = nodes[j];
            if (!child.IsChild) { // must copy complete subtree
              Array.Copy(nodes, j - child.Size, sorted, idx, child.Size);
              idx += child.Size;
            }
            sorted[idx++] = nodes[j];
          }
          Array.Copy(sorted, 0, nodes, i - size, size);
        }
        node.Hash = ComputeHash(ha, nodes, i);
      }
    }

    /// <summary>
    /// Get a function node's child indices from left to right
    /// </summary>
    /// <param name="nodes"></param>
    /// <param name="i"></param>
    /// <returns>An array containing child indices</returns>
    private static int[] IterateChildren(Node[] nodes, int i) {
      var node = nodes[i];
      var arity = node.Arity;
      var children = new int[arity];
      var idx = i - 1;
      for (int j = 0; j < arity; ++j) {
        children[j] = idx;
        idx -= 1 + nodes[idx].Size;
      }
      return children;
    }

    private static void UpdateNodeSizes(Node[] nodes) {
      for (int i = 0; i < nodes.Length; ++i) {
        var node = nodes[i];
        if (node.IsChild) {
          node.Size = 0;
          continue;
        }
        node.Size = node.Arity;
        foreach (var j in IterateChildren(nodes, i)) {
          node.Size += nodes[j].Size;
        }
      }
    }

    /// <summary>
    /// Reduce nodes of the same type according to arithmetic rules
    /// </summary>
    /// <param name="g">The given grammar (controls symbol arities)</param>
    /// <param name="nodes">The node array</param>
    /// <param name="i">Parent index</param>
    /// <param name="j">Child index</param>
    private static void Reduce(this Grammar g, Node[] nodes, int i, int j) {
      var p = nodes[i]; // parent node
      var c = nodes[j]; // child node

      if (c.IsChild)
        return;

      foreach (var k in IterateChildren(nodes, j)) {
        if (nodes[k].IsChild) continue;
        g.Reduce(nodes, j, k);
      }

      // handle commutative symbols (add, mul) 
      if (g.Arity(p.Symbol) == 2 && p.Symbol == c.Symbol) {
        c.Enabled = false;
        p.Arity += c.Arity - 1;
      }
    }

    private static Node[] Reduce(this Grammar grammar, Node[] nodes) {
      UpdateNodeSizes(nodes);
      int i = nodes.Length - 1;
      foreach (var c in IterateChildren(nodes, i)) {
        if (nodes[c].IsChild) continue;
        grammar.Reduce(nodes, i, c);
      }
      int count = 0;
      foreach (var node in nodes) {
        if (!node.Enabled) { ++count; }
      }
      if (count == 0)
        return nodes;

      var reduced = new Node[nodes.Length - count];
      i = 0;
      foreach (var node in nodes) {
        if (node.Enabled) { reduced[i++] = node; }
      }
      UpdateNodeSizes(reduced);
      return reduced;
    }
  }
}
